#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "s2plot.h"

/* Global variables - these are seen by all functions */
int vid;		/* ID for volume render object */
int iso;		/* ID for isosurface object */
float ***volume;	/* The data cube */
int N = -1;		/* Number of data points: default for error checking */
COLOUR *col;		/* Array of colour values - size to be set */
XYZ *xyz;		/* Array of data values - size to be set */

float ***initVolume(int nx, int ny, int nz)
/* Allocate memory and initialise a data cube */
{
   float ***volume;
   int i, j, k;

   volume = (float ***)malloc(nx * sizeof(float **));
   if (volume == NULL) {
      fprintf(stderr,"Failed to allocate %ld bytes\n",nx*sizeof(float **));
      exit(-1);
   }
   for (i=0;i<nx;i++) {
      volume[i] = (float **)malloc(ny * sizeof(float *));
      if (volume[i] == NULL) {
         fprintf(stderr,"Failed to allocate %ld bytes\n",nx*sizeof(float *));
         exit(-1);
      }
      for (j=0;j<ny;j++) {
         volume[i][j] = (float *)malloc(nz * sizeof(float));
         if (volume[i][j] == NULL) {
            fprintf(stderr,"Failed to allocate %ld bytes\n",nx*sizeof(float));
            exit(-1);
         }
         for (k=0;k<nz;k++) {
            volume[i][j][k] = 0.0;		/* Initialise */
         }
      }
   }

   return volume;
}


int countLines(FILE *fp)
/* Return the number of lines in the file */
{
   int N = 0;			/* Count of number of lines */
   long int ptr;		/* Location within file */
   char string[1024];		/* Line read from file */

   ptr = ftell(fp);       	/* Current location in file */
   
   fgets(string,1024,fp);	/* Read some data */
   while (!feof(fp)) { fgets(string,1024,fp); N++; }	/* Add line to counter */
   fseek(fp, ptr, SEEK_SET);            /* Back to where file was */

   return N;			/* Return number of lines */
}

float smooth(float x1, float x2, float y1, float y2, float z1, float z2,
                int M, XYZ *xyz, int N, float ***grid)
/* Smooth N (x, y, z) values onto a grid with dimensions M*M*M using */
/* the triangular-shaped cloud algorithm */
{
   int i, j, k, m;
   XYZ cen, diff;
   float dx, dy, dz;
   float wx, wy, wz, w;
   float ddx, ddy, ddz;
   float cw = 0;
   int match;
   int jmin,jmax;
   int kmin,kmax;
   int mmin,mmax;
   int gx,gy,gz;

   dx = (x2-x1)/M;
   dy = (y2-y1)/M;
   dz = (z2-z1)/M;
   match = 0;
   for (i=0;i<N;i++) {
      gx = (xyz[i].x-x1)/dx;            /* Guess at which grid cell is centre */
      gy = (xyz[i].y-y1)/dy;
      gz = (xyz[i].z-z1)/dz;
      jmin = gx-2; if (jmin < 0) jmin = 0;      /* Only search near this */
      jmax = gx+2; if (jmax > M) jmax = M;
      kmin = gy-2; if (kmin < 0) kmin = 0;
      kmax = gy+2; if (kmax > M) kmax = M;
      mmin = gz-2; if (mmin < 0) mmin = 0;
      mmax = gz+2; if (mmax > M) mmax = M;
      for (j=jmin;j<jmax;j++) {
         cen.x = (j+0.5)*dx + x1;               /* Cell centre: x */
         diff.x = fabs(xyz[i].x - cen.x);       /* Distance to cell centre: x */
         wx = 0.0;                              /* Reset weight */
         ddx = diff.x/dx;
         if (ddx <= 0.5) { wx = 0.75 - ddx*ddx; }
         else if ((ddx > 0.5) && (ddx <= 1.5))  /* Weight conditions */
                { wx = 0.5*(1.5 - ddx)*(1.5-ddx); }
         for (k=kmin;k<kmax;k++) {
            cen.y = (k+0.5)*dy + y1;            /* Cell centre: y */
            diff.y = fabs(xyz[i].y - cen.y);    /* Distance to cell centre: y */
            wy = 0;                             /* Reset weight */
            ddy = diff.y/dy;
            if (ddy <= 0.5) { wy = 0.75 - ddy*ddy; }
            if ((ddy > 0.5) && (ddy <= 1.5))    /* Weight condition */
                { wy = 0.5*(1.5 - ddy)*(1.5-ddy); }
            for (m=mmin;m<mmax;m++) {
               cen.z = (m+0.5)*dz + z1;         /* Cell centre: z */
               diff.z = fabs(xyz[i].z - cen.z); /* Distance to cell centre: z */
               wz = 0;                          /* Reset weight */
               ddz = diff.z/dz;
               if (ddz <= 0.5) { wz = 0.75 - ddz*ddz; }
               if ((ddz > 0.5) && (ddz <= 1.5))         /* Weight condition */
                { wz = 0.5*(1.5 - ddz)*(1.5-ddz); }
               w = wx*wy*wz;
               if (w) {
                  if (((j-gx) == 0) && ((k-gy) == 0) & ((m-gz) == 0)) {
                                                /* Was something found */
                      match++;
                  }
                  grid[j][k][m] += w;
                  cw += w;
              }
            }
         }
      }
   }
   if (match != N) { fprintf(stderr,"TSC missed %d\n", N-match); }
   return cw;
}


int readData(char *input)
/* Read dat from the input file */
{
   FILE *fp;				/* Pointer to input file */
   char string[256];			/* Text read from file */
   int N;				/* Number of particles from file */
   int i;				/* Loop variable */

/* Open the data file and check that it did open */
   if ((fp = fopen(input,"r")) == NULL) {	
      fprintf(stderr,"Could not open data file: %s\n", input);
      exit(0);
   }

   N = countLines(fp);				/* Count number of particles */

   xyz = (XYZ *)calloc(N, sizeof(XYZ));		/* Allocate global array mem */
   col = (COLOUR *)calloc(N, sizeof(COLOUR)); 	/* Allocate global array mem */

   for (i=0;i<N;i++) {
      fgets(string, 255, fp);	/* Get a line of text from file */
      sscanf(string,"p %lf %lf %lf %lf %lf %lf",&xyz[i].x, &xyz[i].y, &xyz[i].z,
		&col[i].r, &col[i].g, &col[i].b);

      xyz[i].x -= 0.5;		/* Shift data so centred on origin */
      xyz[i].y -= 0.5;
      xyz[i].z -= 0.5;
   }
   return N;	/* Return number of lines read */
}


void cb(double *t, int *kc)
/* A dynamic callback */
{
   int i;
   if (*kc % 3 == 0) {
      for(i=0;i<N;i++) {		/* Draw the points */
         ns2vpoint(xyz[i], col[i]);
      }
   } else if (*kc % 3 == 1) {
      ds2dvr(vid, 1);			/* Draw the volume render object */
      for(i=0;i<N;i++) {		/* Draw the points */
         ns2vpoint(xyz[i], col[i]);
      }
   } else if (*kc % 3 == 2) {
      ns2dis(iso, 1);			/* Draw the volume render object */
      for(i=0;i<N;i++) {		/* Draw the points */
         ns2vpoint(xyz[i], col[i]);
      }
   }
}

int main(int argc, char *argv[])
{
   int i, j, k; 			/* Loop variable */
   int nx = 40, ny = 40, nz = 40;	/* Dimensions of data cube */ 
   float dmin, dmax, amin, amax;	/* Min/max values */
   char trans;				/* Transparency mode */
   float tr[12];			/* Transformation matrix */
   float x1 = -0.25, x2 = 0.25;		/* Data range */
   float y1 = -0.25, y2 = 0.25;		/* Data range */
   float z1 = -0.25, z2 = 0.25;		/* Data range */
   float dx, dy, dz;			/* Temporary variables */
   COLOUR amb = {0.8, 0.8, 0.8};	/* Ambient light */
   float slope;				/* Parameter controlling volume scale */

   volume = initVolume(nx, ny, nz); 	/* Allocate memory for data volume */  

   for (i=0;i<12;i++) {			/* Set-up transfrom matrix */
      tr[i] = 0.0;
   }
   tr[ 0]  = x1; 			/* Mapping from data cube */
   tr[ 1]  = (x2-x1)/(float)(nx-1.0);	/* to physical coordinates */
   tr[ 4]  = y1; 
   tr[ 6]  = (y2-y1)/(float)(ny-1.0);
   tr[ 8]  = z1; 
   tr[11]  = (z2-z1)/(float)(nz-1.0);

   dx = tr[1]*0.5;		/* Offsets for window - voxels are pixel */
   dy = tr[6]*0.5;		/*  centred */
   dz = tr[11]*0.5;

   N = readData("grey.geom");	/* Read the input data */

   smooth(x1, x2, y1, y2, z1, z2, nx, xyz, N, volume);	/* Smooth onto grid */

/*** TRY CHANGING THE VALUE OF THE SLOPE PARAMETER ***/
   slope = 0.5;			/* Volume scaling */

/* Find the maximum and minimum data value - required for volume render */
   dmin = +1E10;
   dmax = -1E10;
   for (i=0;i<nx;i++) {
      for (j=0;j<ny;j++) {
         for (k=0;k<nz;k++) {
            volume[i][j][k] = pow(volume[i][j][k], slope);	/* Rescale data */
            if (volume[i][j][k] < dmin) dmin = volume[i][j][k];
            if (volume[i][j][k] > dmax) dmax = volume[i][j][k];
         }
      }
   }
 
   dmax *= 0.9;		/* Bring down the peak value slightly */

   s2opend("/?",argc,argv);				/* Open the display */
   s2swin(x1-dx,x2+dx, y1-dy,y2+dy, z1-dz,z2+dz);	/* Set window coords */
   s2box("BCDE",0,0,"BCDE",0,0,"BCDE",0,0);		/* Draw coord box */

   s2scir(100,200);			/* Set colour range */

/*** TRY SOME OTHER COLOUR MAPS ***/
/*** "iron", "inverse iron", "terrain", "mgreen" ***/

   s2icm("rainbow",100,200);		/* Install colour map */

   amin = 0.0;			/* Minimum alpha channel value */
   amax = 0.7;			/* Maximum alpha channel value */
   trans = 't';			/* Transparency type */

   vid = ns2cvr(volume, nx, ny, nz, 0, nx-1, 0, ny-1, 0, nz-1, 
		tr, trans, dmin, dmax, amin, amax);
					/* Create the volume render object */

/*** CHANGE THE VALUE OF LEVEL TO SEE EFFECT OF DIFFERENT ISOSURFACE ***/
   float level = 0.15*dmax;


   iso = ns2cis(volume, nx, ny, nz, 0, nx-1, 0, ny-1, 0, nz-1, 
		tr, level, 1, trans, 0.2, 1.0, 1.0, 0.0);
					/* Create the isosurface object */

   s2scb(cb);				/* Install a dynamic callback */

   ss2srm(SHADE_FLAT);			/* Set shading type to FLAT */
   ss2sl(amb, 0, NULL, NULL, 0);	/* Ambient lighting only */

   s2show(1);				/* Open the s2plot window */

   return 1;
}
